import { world, system, ItemStack } from "@minecraft/server";

const suits = ["spades", "hearts", "diamonds", "clubs"];
const ranks = ["ace", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten", "jack", "queen", "king"];
const entityCardSet = new Set(
  suits.flatMap(s => ranks.map(r => `gpc:${r}_of_${s}`))
    .concat(["gpc:black_joker", "gpc:red_joker"])
);
const cardSet = new Set(
  suits.flatMap(s => ranks.map(r => `gpc:${r}_of_${s}_spawn_egg`))
    .concat(["gpc:black_joker_spawn_egg", "gpc:red_joker_spawn_egg"])
);
const overworld = world.getDimension("overworld");
const lastCounts = new Map();
const pemainTerdeteksi = new Set();
let sudahMenang = false, sudahPenalti = false;

const isNearby = (a, b, d = 10) => (a.x - b.x) ** 2 + (a.y - b.y) ** 2 + (a.z - b.z) ** 2 <= d * d;
const sayChange = (p, d, c) => `tellraw @a[r=10] {"rawtext":[{"text":"§a${p.name}§r "},{"translate":"gpc.change","with":["${d > 0 ? "+" : "-"}${Math.abs(d)}","${Math.abs(d) === 1 ? "card" : "cards"}","${c}","${c === 1 ? "Card" : "Cards"}"]}]}`;

function lockAllPlayerCards(player, cardSet) {
  if (!player.hasTag("peserta_uno")) return;

  const inv = player.getComponent("inventory")?.container;
  if (!inv) return;

  const slotsToReplace = [];

  for (let i = 0; i < 9; i++) {
    const item = inv.getItem(i);
    if (item && cardSet.has(item.typeId)) {
      slotsToReplace.push({
        slot: i,
        typeId: item.typeId,
        amount: item.amount,
      });
    }
  }

  const totalCards = slotsToReplace.reduce((sum, entry) => sum + entry.amount, 0);
  const isLocked = player.hasTag("locked");

  if (totalCards <= 5) {
    for (const entry of slotsToReplace) {
      player.runCommandAsync(
        `replaceitem entity "${player.name}" slot.hotbar ${entry.slot} ${entry.typeId} ${entry.amount} 0 {"minecraft:item_lock":{"mode":"lock_in_slot"}}`
      );
    }
    if (!isLocked) player.addTag("locked");
  } else if (totalCards > 5 && isLocked) {
    for (let i = 0; i < 9; i++) {
      const item = inv.getItem(i);
      if (item && cardSet.has(item.typeId)) {
        player.runCommandAsync(
          `replaceitem entity "${player.name}" slot.hotbar ${i} ${item.typeId} ${item.amount} 0`
        );
      }
    }
    player.removeTag("locked");
  }
}

let sudahUno = false;

system.runInterval(() => {
  let adaPemenang = false, adaPenalti = false;
  const dealerEntities = overworld.getEntities({ type: "gpc:playing_card_box", tags: ["uno_dealer"] });

  const dealer = dealerEntities[0];
  if (!dealer) return;

  for (const player of world.getPlayers()) {
    const inv = player.getComponent("inventory")?.container;
    if (!inv) continue;

    const invItems = [...Array(inv.size)].map((_, i) => inv.getItem(i));
    const jumlah = invItems.reduce((t, item) => item && cardSet.has(item.typeId) ? t + item.amount : t, 0);

    const dealerNearby = dealerEntities.find(d => isNearby(player.location, d.location));
    if (!dealerNearby) continue;

    const dealerName = dealerNearby.nameTag;
    const has = t => player.hasTag(t);
    const add = t => !has(t) && player.addTag(t);
    const remove = t => has(t) && player.removeTag(t);

    if (jumlah > 0 && has("draw_dealer")) add("peserta_uno");

    if (dealerName === "Uno") {
      if (has("peserta_uno") && has("uno") && jumlah === 0) add("menang");
      if (has("peserta_uno") && has("uno") && jumlah >= 3) remove("uno");
      if (has("peserta_uno") && !has("menang") && !has("uno") && jumlah === 1 && !has("penalti")) {
        add("penalti");
        adaPenalti = true;
      } else if (has("penalti") && jumlah !== 1) {
        remove("penalti");
      }
    }

    else if (dealerName === "Cangkul") {
      if (has("peserta_uno") && jumlah === 0) add("menang");
    }

    if (has("menang")) adaPemenang = true;

    for (let i = 0; i < 9; i++) {
      if (!invItems[i]) {
        const idx = invItems.findIndex((item, j) => j >= 9 && item && cardSet.has(item.typeId));
        if (idx !== -1) {
          inv.setItem(i, invItems[idx]);
          inv.setItem(idx, undefined);
          invItems[i] = invItems[idx];
          invItems[idx] = undefined;
        }
      }
    }

    const last = lastCounts.get(player.name) ?? jumlah;
    if ((has("draw_dealer") || has("peserta_uno")) && jumlah > last) {
      dealer.runCommandAsync(sayChange(player, jumlah - last, jumlah));

      const inv = player.getComponent("inventory")?.container;
      const hasSpyglass = [...Array(inv.size)].some((_, i) => {
        const item = inv.getItem(i);
        return item?.typeId === "minecraft:spyglass";
      });

      if (!hasSpyglass) {
        dealer.runCommandAsync(`give "${player.name}" spyglass`);
      }

      player.removeTag("draw_dealer");

      if (has("uno") || has("pencet_uno")) {
        player.removeTag("uno");
        player.removeTag("pencet_uno");
      }

      lockAllPlayerCards(player, cardSet);
    }

    if (has("peserta_uno") && jumlah < last) {
      dealer.runCommandAsync(sayChange(player, jumlah - last, jumlah));
      lockAllPlayerCards(player, cardSet);
    }
    lastCounts.set(player.name, jumlah);
  }

  const pencetUno = world.getPlayers().find(p => p.hasTag("peserta_uno") && p.hasTag("uno") && !p.hasTag("pencet_uno"));
  if (pencetUno && !sudahUno) {
    dealer.runCommandAsync(`title @a[r=16] title §a${pencetUno.name} Uno!§r`);
    dealer.runCommandAsync(`tellraw @a[r=16] {"rawtext":[{"text":"§a${pencetUno.name} Uno!§r"}]}`);
    dealer.runCommandAsync(`playsound uno.uno @a ~~~ 0.75`);
    pencetUno.addTag("pencet_uno");
    sudahUno = true;
  } else if (!pencetUno) {
    sudahUno = false;
  }

  if (adaPemenang && !sudahMenang) {
    const pemenang = world.getPlayers().find(p =>
      p.hasTag("menang") && p.hasTag("peserta_uno")
    );
    if (pemenang) {
      pemenang.removeTag("peserta_uno");
      pemenang.removeTag("uno");
      pemenang.removeTag("pencet_uno");
      pemenang.removeTag("locked");
      dealer.runCommandAsync(`playsound uno.win @a ~~~ 0.75`);
	  dealer.runCommandAsync(`titleraw @a[r=16] title {"rawtext":[{"translate":"gpc.win.title","with":["${pemenang.name}"]}]}`);
	  dealer.runCommandAsync(`tellraw @a[r=16] {"rawtext":[{"translate":"gpc.win.say","with":["${pemenang.name}"]}]}`);
      pemenang.removeTag("menang");
      sudahMenang = true;
    }
  } else if (!adaPemenang) {
    sudahMenang = false;
  }

  const pemainPenalti = world.getPlayers().find(p =>
    p.hasTag("penalti") && p.hasTag("peserta_uno")
  );
  if (adaPenalti && !sudahPenalti) {
    pemainPenalti.addTag("draw_dealer");
	dealer.runCommandAsync(`tellraw @a[r=16] {"rawtext":[{"translate":"gpc.penalty.say","with":["${pemainPenalti.name}"]}]}`);
	dealer.runCommandAsync(`titleraw @a[r=16] title {"rawtext":[{"translate":"gpc.penalty.title","with":["${pemainPenalti.name}"]}]}`);
	dealer.runCommandAsync(`titleraw @a[r=16] subtitle {"rawtext":[{"translate":"gpc.penalty.subtitle"}]}`);
    dealer.runCommandAsync(`playsound uno.lupa_uno @a ~~~ 0.75`);
    dealer.runCommandAsync(`execute as @e[type=gpc:deck,c=1] at @s run tag @r[family=card,type=!gpc:deck,c=2,x=~,y=~-1,z=~,dx=0,dy=2,dz=0] add draw_card`);
    dealer.runCommandAsync(`effect @e[tag=draw_card] invisibility infinite 255 true`);
    dealer.runCommandAsync(`kill @e[tag=draw_card]`);
    sudahPenalti = true;
  } else if (!adaPenalti) {
    sudahPenalti = false;
  }

  dealer.runCommandAsync(`execute as @a[tag=peserta_uno,hasitem={item=spyglass,quantity=1,location=slot.weapon.mainhand}] at @e[type=gpc:playing_card_box,tag=uno_dealer,c=1] run camera @s[r=2.5] set minecraft:free pos ~ ~-0.06 ~ rot ~90 ~-90`);
  dealer.runCommandAsync(`execute as @a[tag=peserta_uno,hasitem={item=spyglass,quantity=1,location=slot.weapon.mainhand}] at @e[type=gpc:playing_card_box,tag=uno_dealer,c=1] run camera @s[rm=2.75] clear`);
  dealer.runCommandAsync(`execute as @a[tag=peserta_uno,hasitem={item=spyglass,quantity=0,location=slot.weapon.mainhand}] run camera @s clear`);
  dealer.runCommandAsync(`execute as @a[tag=!peserta_uno,hasitem={item=spyglass,quantity=1,location=slot.weapon.mainhand},r=5] at @e[type=gpc:playing_card_box,tag=uno_dealer,c=1] run camera @s set minecraft:free pos ~-3 ~2 ~ rot ~44 ~-90`);
  dealer.runCommandAsync(`execute as @a[tag=!peserta_uno,hasitem={item=spyglass,quantity=1,location=slot.weapon.mainhand},rm=5.25] at @e[type=gpc:playing_card_box,tag=uno_dealer,c=1] run camera @s clear`);
  dealer.runCommandAsync(`execute as @a[tag=!peserta_uno,hasitem={item=spyglass,quantity=0,location=slot.weapon.mainhand}] at @e[type=gpc:playing_card_box,tag=uno_dealer,c=1] run camera @s clear`);

  const players = world.getPlayers().filter(p => p.hasTag("draw_dealer") && dealerEntities.some(d => isNearby(p.location, d.location)));
  const eliminated = new Set();

  for (const shuffler of overworld.getEntities({ type: "gpc:deck" })) {
    const items = overworld.getEntities({
      type: "minecraft:item",
      location: shuffler.location,
      maxDistance: 2
    }).filter(item => {
      const stack = item.getComponent("minecraft:item")?.itemStack;
      return stack && cardSet.has(stack.typeId);
    });

    for (const item of items) {
      const stack = item.getComponent("minecraft:item")?.itemStack;
      if (!stack) continue;

      for (const player of players) {
        const inv = player.getComponent("inventory")?.container;
        if (!inv) continue;

        const invItems = [...Array(inv.size)].map((_, i) => inv.getItem(i));
        const isFull = invItems.every(item => item && item.amount > 0);

        if (isFull) {
          if (!player.hasTag("elim_full")) player.addTag("elim_full");
        } else {
          if (inv.addItem(stack)) item.kill();
          else item.kill();
        }

        break;
      }
    }
  }

  for (const player of world.getPlayers()) {
    if (player.hasTag("elim_full") && !eliminated.has(player.name)) {
      if (dealer) {
        const nearbyItems = overworld.getEntities({
          type: "minecraft:item",
          location: dealer.location,
          maxDistance: 10
        });

        for (const item of nearbyItems) {
          const stack = item.getComponent("minecraft:item")?.itemStack;
          if (stack && cardSet.has(stack.typeId)) item.kill();
        }

        dealer.runCommandAsync(`tp "${player.name}" ~-13 ~-2 ~ facing ~-12 ~-2 ~`);
        const tagsToRemove = ["peserta_uno", "draw_dealer", "uno", "menang", "penalti", "elim_full"];
        for (const tag of player.getTags()) {
          if (tagsToRemove.includes(tag)) {
            player.removeTag(tag);
          }
        }
        dealer.runCommandAsync(`execute as "${player.name}" at @s run function uno/clear_cards`);
        dealer.runCommandAsync(`tellraw @a[r=16] {"rawtext":[{"translate":"gpc.kick.inv.say","with":["${player.name}"]}]}`);
		dealer.runCommandAsync(`titleraw @a[r=16] title {"rawtext":[{"translate":"gpc.kick.inv.title","with":["${player.name}"]}]}`);
		dealer.runCommandAsync(`titleraw @a[r=16] subtitle {"rawtext":[{"translate":"gpc.kick.inv.subtitle"}]}`);
        dealer.runCommandAsync(`execute as @a at @s run playsound mob.elderguardian.curse @s ~~~ 0.75`);
      }
      eliminated.add(player.name);
    }
  }
}, 1);

system.runInterval(() => {
  const dealer = overworld.getEntities({ type: "gpc:playing_card_box", tags: ["uno_dealer"] }).find(entity => entity);
  if (!dealer) return;
  const bundleItems = ["minecraft:bundle", ...["red", "blue", "green", "yellow", "black", "white", "purple", "cyan", "magenta", "orange", "brown", "gray", "light_gray", "lime", "pink", "light_blue"].map(color => `minecraft:${color}_bundle`)];

  for (const player of world.getPlayers()) {
    if (!player.hasTag("peserta_uno") || player.hasTag("elim_full")) continue;

    const dx = player.location.x - dealer.location.x;
    const dy = player.location.y - dealer.location.y;
    const dz = player.location.z - dealer.location.z;
    const distanceSq = dx * dx + dy * dy + dz * dz;

    if (distanceSq <= 36) {
      const inventory = player.getComponent("inventory").container;
      let hasBundle = false;

      for (let i = 0; i < inventory.size; i++) {
        const item = inventory.getItem(i);
        if (item && bundleItems.includes(item.typeId)) {
          hasBundle = true;
          break;
        }
      }

      if (hasBundle) {
        player.addTag("elim_full");
        const playerName = player.name;
        dealer.runCommandAsync(`tellraw @a[r=16] {"rawtext":[{"translate":"gpc.kick.bundle","with":["${playerName}"]}]}`);
      }
    }
  }
}, 1);

const shufflerState = new WeakMap();

system.runInterval(() => {
  const deck = world.getDimension("overworld").getEntities({ type: "gpc:deck" })[0];

  if (deck) {
    const nearbyCards = deck.dimension.getEntities({
      location: deck.location,
      maxDistance: 2,
      families: ["card"],
      excludeTypes: ["gpc:deck"],
    });

    const cardCount = nearbyCards.length;
    const state = shufflerState.get(deck) ?? false;

    if (cardCount <= 15 && !state) {
      const dealer = world.getDimension("overworld").getEntities({
        type: "gpc:playing_card_box",
        tags: ["uno_dealer"]
      })[0];

      if (!dealer) return;

      dealer.runCommandAsync(`tellraw @a[r=16] {"rawtext":[{"translate":"gpc.refill.say"}]}`);
	  dealer.runCommandAsync(`titleraw @a[r=16] title {"rawtext":[{"translate":"gpc.refill.title"}]}`);
	  dealer.runCommandAsync(`titleraw @a[r=16] subtitle {"rawtext":[{"translate":"gpc.refill.subtitle"}]}`);
      dealer.runCommandAsync(`playsound random.toast @a ~~~ 1`);

      shufflerState.set(deck, true);
    }

    if (cardCount > 15 && state) {
      shufflerState.set(deck, false);
    }
  }
}, 60);

system.runTimeout(() => {
  system.runInterval(() => {
    const onlineNow = new Map();
    for (const p of world.getPlayers()) {
      onlineNow.set(p.id, p);
    }

    for (const player of onlineNow.values()) {
      if (!pemainTerdeteksi.has(player.id)) {
        pemainTerdeteksi.add(player.id);
        const playerId = player.id;
        system.runTimeout(() => {
          const stillOnline = world.getPlayers().find(p => p.id === playerId);
          if (!stillOnline) return;
          stillOnline.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§ePlaying Cards addon by §aPapa Alsa§r"}]}`);
		  stillOnline.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cYouTube: @alsahonggo §r"}]}`);
		  stillOnline.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§e-----"}]}`);
          stillOnline.runCommandAsync(`tellraw @s {"rawtext":[{"translate":"gpc.setup"}]}`);
          stillOnline.runCommandAsync("playsound random.toast @s ~~~ 1");
        }, 200);
      }
    }

    for (const id of pemainTerdeteksi) {
      if (!onlineNow.has(id)) pemainTerdeteksi.delete(id);
    }
  }, 20);
}, 20);

world.beforeEvents.worldInitialize.subscribe(e => {
  e.blockComponentRegistry.registerCustomComponent('gpc:duduk_uno', {
    onPlayerInteract: e => {
      const { block, player } = e;
      if (block?.typeId === "gpc:kursi_uno") {
        const loc = { ...block.center(), y: block.center().y - 0.5 };
        if (block.dimension.getEntities({ location: loc, type: "gpc:kursi_uno_entity", maxDistance: 0.3 }).length === 0) {
          block.dimension.spawnEntity("gpc:kursi_uno_entity", loc).getComponent("rideable")?.addRider(player);
        }
      }
    }
  });
});

world.afterEvents.playerBreakBlock.subscribe(e => {
  if (e.brokenBlockPermutation.type.id === "gpc:kursi_uno") {
    const { x, y, z } = e.block.location;
    e.block.dimension.getEntities({ location: { x, y, z }, type: "gpc:kursi_uno_entity" })
      .forEach(ent => {
        const pos = ent.location;
        if (Math.floor(pos.x) === x && Math.floor(pos.y) === y && Math.floor(pos.z) === z) ent.remove();
      });
  }

  const dealerEntities = overworld.getEntities({ type: "gpc:playing_card_box", tags: ["uno_dealer"] });
  const dealer = dealerEntities[0];
  if (!dealer) return;
  const blockType = e.brokenBlockPermutation?.type?.id;
  if (!blockType || (!blockType.includes("carpet") && blockType !== "minecraft:sea_lantern")) return;

  const dx = dealer.location.x - e.block.location.x;
  const dy = dealer.location.y - e.block.location.y;
  const dz = dealer.location.z - e.block.location.z;
  const distSq = dx * dx + dy * dy + dz * dz;
  if (distSq <= 12) {
    overworld.getBlock(e.block.location)?.setType(e.brokenBlockPermutation.type);

    const droppedItems = overworld.getEntities({
      type: "minecraft:item",
      location: e.block.location,
      maxDistance: 3
    });

    for (const item of droppedItems) {
      const itemComp = item.getComponent("minecraft:item");
      if (!itemComp) continue;

      const itemId = itemComp.itemStack?.typeId;
      if (itemId && (itemId.includes("carpet") || itemId === "minecraft:prismarine_crystals")) {
        item.kill();
      }
    }
  }
});

world.afterEvents.entityDie.subscribe(event => {
  const entity = event.deadEntity;
  const killer = event.damageSource?.damagingEntity;

  if (!entityCardSet.has(entity.typeId)) return;

  const dropItemId = `${entity.typeId}_spawn_egg`;

  if (!cardSet.has(dropItemId)) {
    return;
  }

  const dropItem = new ItemStack(dropItemId);
  dropItem.amount = 1;

  if (killer?.typeId === "minecraft:player") {
    const inv = killer.getComponent("inventory")?.container;
    if (!inv) return;
    inv.addItem(dropItem);

    let found = false;
    for (let i = 0; i < inv.size; i++) {
      const item = inv.getItem(i);
      if (item?.typeId === dropItemId) {
        found = true;
        break;
      }
    }

    if (found) {
    } else {
      entity.dimension.spawnItem(dropItem, entity.location);
    }

  } else {
    entity.dimension.spawnItem(dropItem, entity.location);
  }
});